from __future__ import annotations
from typing import Dict, List, Union, Any
import logging
from pydoc import locate

from attrdict import AttrDict
from flask_sqlalchemy import Model
from sqlalchemy import Column

from api.models import *  # noqa

logger = logging.getLogger(__name__)


class Endpoint(object):
    def __init__(
        self,
        name: str,
        model: str,
        version: str = None,
        path: str = "/",
        mappings: Dict[str, str] = None,
        url_params: List[str] = None,
        exclude: List[str] = None,
        depends_on: Dict[str, str] = None,
        options: List[str] = None,
        normalize: bool = False,
        updated_column: str = "updated_at",
        enabled: bool = True,
        ignore_unknown: bool = True,
        **kwargs,
    ):

        self.name = name
        self.version = version
        self.path = path
        self.mappings = AttrDict(mappings or {})
        self.model_name = model
        self._model = None
        self.updated_column = updated_column
        self.url_params = url_params
        self._exclude = exclude or []
        self._dependency_map = depends_on or {}
        self._depends_on: Union[Dict[str, Model], None] = None
        self.normalize = normalize
        self.options = options or []
        self.enabled = enabled
        self.ignore_unknown = ignore_unknown

    def __repr__(self):
        return self.path

    def __iter__(self):
        attrs = [x for x in dir(self) if not x.startswith("_")]
        for a in attrs:
            yield a, getattr(self, a)

    @property
    def model(self) -> Model:
        if self._model is None:
            self._model = self.locate_model(self.model_name)
        return self._model

    @property
    def depends_on(self) -> Dict[str, Column]:
        if self._depends_on is None:
            self._depends_on = self.link_upstream_columns()
        return self._depends_on

    @property
    def exclude(self):
        return self._exclude

    @property
    def known_columns(self):
        return self._exclude + self.mapped_names

    @property
    def alias_map(self) -> Dict[str, str]:
        return self.mappings.get("aliases", {})

    @property
    def mapped_names(self) -> List[str]:
        return list(self.alias_map.keys())

    @property
    def mapped_aliases(self) -> List[str]:
        return list(self.alias_map.values())

    def link_upstream_columns(self) -> Dict[str, Column]:
        linked_columns: Dict[Any, Any] = {}
        for key_field, model_name in self._dependency_map.items():
            model_name, column_name = model_name.rsplit(".", maxsplit=1)
            columns = list(self.locate_model(model_name).__table__.columns)
            column_map = {c.name: c for c in columns}
            if linked_columns.get(key_field) is None:
                linked_columns[key_field] = column_map[column_name]
            else:
                linked_columns[key_field] += column_map[column_name]
        return linked_columns

    def locate_model(self, model_name: str) -> Model:
        model: Model = None
        try:
            # try to import dotted model name. ex: api.models.MyModel
            model = locate(model_name)
            logger.debug(f"Found model: {model_name}")
        except ModuleNotFoundError:
            logger.debug(
                f"Failed to import module '{model_name}' from project directory"
            )
            try:
                # try to import model from global namespace
                model = globals()[model_name]
                logger.debug(f"Model '{model_name}' found in global namespace")
            except ModuleNotFoundError:
                raise ModuleNotFoundError(
                    f"Name '{model_name}' not found in project or global namespace"
                )
        except Exception as e:
            raise Exception(f"Unable to locate module {model_name} -- {e}")

        return model

    @staticmethod
    def load_from_config(
        app_config: object, load_disabled: bool = False
    ) -> Dict[str, Endpoint]:
        endpoints: dict = {}
        try:
            endpoints = app_config.endpoints  # type: ignore
        except AttributeError:
            raise AttributeError("Config object has no attribute 'endpoints'")

        loaded: Dict[str, Endpoint] = {}
        for ep in endpoints.items():
            try:
                new = Endpoint(name=ep[0], **ep[1])
                if new.enabled or load_disabled:
                    loaded[ep[0]] = new
                    # print(f"Created endpoint ({ep[0]})")
                # else:
                #     print(f"Skipping endpoint ({ep[0]})")
            except Exception as e:
                logger.error(f"Failed to create endpoint ({ep[0]}) -> {e}")

        return loaded

    @staticmethod
    def from_dict(name: str, data: dict):
        return Endpoint(name, **data)


if __name__ == "__main__":

    from config import get_active_config

    conf = get_active_config()
    endpoints = conf.endpoints
    # [x for x in endpoints.items()]
    Endpoint.load_from_config(conf)

    # e = Endpoint("test", **{"model": "test"})
    e = Endpoint(name="test", **endpoints["frac_schedules"])
    e.exclude



from __future__ import annotations
from typing import Union, Dict, List, Any
from datetime import datetime, timedelta
import os
import tempfile
from contextlib import contextmanager
import logging
from collections import Counter

import yaml

logger = logging.getLogger(__name__)


class Yammler(dict):
    _no_dump = ["changed"]
    _metavars = ["fspath", "updated_at"]
    _data_key = "data"
    _meta_key = "meta"

    def __init__(self, fspath: str, data: dict = None):
        self.fspath = fspath
        self.changed = False
        self.updated_at = self.stamp()
        data = {}
        if os.path.exists(fspath):
            with open(fspath) as f:
                data = yaml.safe_load(f) or {}
        super().__init__(data)

    def dump(self) -> None:
        with self.durable(self.fspath, "w") as f:
            yaml.safe_dump(dict(self), f, default_flow_style=False)

    @staticmethod
    def stamp():
        return datetime.utcnow()

    @classmethod
    @contextmanager
    def context(cls, fspath):
        obj = cls(fspath)
        try:
            yield obj
        finally:
            obj.dump(force=True)

    @classmethod
    @contextmanager
    def durable(cls, fspath: str, mode: str = "w+b"):
        """ Safely write to file """
        _fspath = fspath
        _mode = mode
        _file = tempfile.NamedTemporaryFile(_mode, delete=False)

        try:
            yield _file
        except Exception as e:  # noqa
            os.unlink(_file.name)
            raise e
        else:
            _file.close()
            os.rename(_file.name, _fspath)




from __future__ import annotations

import io
import logging
import os
import ssl
from ftplib import FTP, FTP_TLS, error_perm
from timeit import default_timer as timer
from typing import Dict, Generator, List, Union

from config import get_active_config
from util import hf_size


logger = logging.getLogger(__name__)

conf = get_active_config()


class ImplicitFTP_TLS(FTP_TLS):
    """FTP_TLS subclass that automatically wraps sockets in SSL to support implicit FTPS.
    Source: https://stackoverflow.com/questions/12164470/python-ftp-implicit-tls-connection-issue"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sock = None

    @property
    def sock(self):
        """Return the socket."""
        return self._sock

    @sock.setter
    def sock(self, value):
        """When modifying the socket, ensure that it is ssl wrapped."""
        if value is not None and not isinstance(value, ssl.SSLSocket):
            value = self.context.wrap_socket(value)
        self._sock = value


class Ftp(FTP):
    """Extends ftplib.FTP, adding components to integrate with an FtpModel
    object to facilitate directory navigation, file listing, and downloading.
    """

    _basepath = "/"

    def __init__(
        self,
        url: str,
        username: str,
        password: str,
        destination: str = None,
        basepath: str = None,
        port: Union[int, str] = 21,
        **kwargs,
    ):
        super().__init__()
        self.url = url
        self.username = username
        self.password = password
        self.destination = destination
        self.connect(host=url, port=int(port))
        self.login(username, password)
        self.m = None  # model hook
        self.basepath = basepath or "."
        self.cwd(self.basepath)

    @property
    def basepath(self):
        return self._basepath

    @basepath.setter
    def basepath(self, path: str) -> None:
        if path == self._basepath:
            logger.info(
                """Attempt to alter basepath resulted in no change. The current
                        value of basepath is equal to the provided value:\n
                        {path}""".format(
                    path=self._basepath
                )
            )
        else:
            try:
                self._basepath = path
                self.cwd(self.basepath)
                logger.debug(f"Basepath changed to {self._basepath}")
            except Exception as e:
                logger.warning(f"{e} -- Could not change basepath to {self._basepath}")

    def list_files(self, path: str = None, contains: str = ".") -> List:
        """Return a listing of the contents in the ftp directory specified by the
        input path. If no path is specified, returns the listing of the ftp's root
        directory.

        Contents of the directory are filtered according to the 'contains' parameter.

        Arguments:
            path {str} -- path do a directory on the ftp server (default: None)
            contains(str) -- string qualifying file names that should be returned
                             (default: '.')

        Returns:
            list -- a list of file names
        """

        return [file for file in self.nlst(path or self.basepath) if contains in file]

    @staticmethod
    def from_config() -> Ftp:
        """Create an Ftp object from environment variables. Looks for environment
           variables prefixed with "FRACX_" that correspond to the required parameter
           name.

           FRACX_FTP_URL = Url to ftp site (e.g. provier.domain.com)
           FRACX_FTP_USERNAME = Username credential
           FRACX_FTP_PASSWORD = Password credential
           FRACX_FTP_PATH = Path to the ftp directory containing the frac schedules
                           (e.g. /example/path/to/FracSchedules/)
           FRACX_DOWNLOAD_DIR = Path to location to save downloaded files


        Returns:
            Ftp
        """
        c = conf.with_prefix("collector_ftp")
        return Ftp(
            url=c.get("url"),
            username=c.get("username"),
            password=c.get("password"),
            basepath=c.get("outpath"),
            port=c.get("port"),
        )

    def upload(self, filename: str):
        """Downloads all files located in the ftp directory located at the basepaths.

        Arguments:
            filename {str} -- path to local file to upload

        Keyword Arguments:
            to {str} -- repository path to which the file should be uploaded. If not
            specified, the current directory of the ftp instance will be used.
            (default: {None})

        """

        to = os.path.join(conf.COLLECTOR_FTP_INPATH, os.path.basename(filename))
        status = "error"
        try:
            with open(os.path.join(filename), "rb") as f:
                print(self.storbinary("STOR " + to, f))
                logger.info("Uploaded: " + filename)

        except error_perm as ep:
            logger.warning(f"{ep} -- {filename}")

        except TimeoutError as te:
            logger.error(f"{te} -- {filename}")
            status = "timeout"
        except Exception as e:
            logger.error(f"{e} -- {filename}")
            status = "error"
        else:
            status = "success"

        return {"to": to, "filename": filename, "status": status}

    def get_all(self) -> Generator:
        """Downloads all files located in the ftp directory located at the basepaths.

        Keyword Arguments:
            to {str} -- local download path. If not specified, the destination path
            used to instatiate the Ftp object is used. If it is also None, the download
            will fail. (default: {None})

        Returns:
            Generator -- returns a generator yielding tuples containing the downloaded
                         file's path, name, and status (e.g. filepath, filename,
                         download_status)

        """

        for filename in self.list_files():
            yield self.get(filename)

    def get(self, filename: Union[str, None]) -> Dict[str, Union[str, bytes, None]]:
        """Implementation of the core downloading function.  Each call to this
        method operates on a single file, downloading it to the designated location.

        Arguments:
            filename {str} -- repository path from which the file should be downloaded

        Keyword Arguments:
            to {str} -- local download path. If not specified, the destination path
            used to instatiate the Ftp object is used. If it is also None, the download
            will fail. (default: {None})

        Returns:
            tuple -- (filepath, filename, download_status)
        """

        status = "ERROR"

        try:
            result = {"status": status, "filename": filename, "content": b""}
            logger.info(f"Starting download from {self.url}...")
            byteio = io.BytesIO()

            ts = timer()
            self.retrbinary("RETR " + filename, byteio.write)  # type: ignore
            te = timer()
            exc_time = round(te - ts, 0)

            content = byteio.getvalue()
            size = len(content)

            logger.info(
                "Download successful (download size: %s, download_time: %ss)",
                hf_size(size or 0),
                exc_time,
                extra={"download_bytes": size, "download_seconds": exc_time},
            )

            status = "success"
            result["content"] = content

        except error_perm as e:
            logger.warning(f"{e} -- {filename}")

        return result

    @property
    def latest_filename(self) -> Union[str, None]:

        names = self.nlst()

        latest_time = None
        latest_name = None

        for name in names:
            time = self.voidcmd("MDTM " + name)
            if (latest_time is None) or (time > latest_time):
                latest_name = name
                latest_time = time

        return latest_name

    def get_latest(self) -> Dict:
        return self.get(self.latest_filename)

    def cleanup(self):
        """ Delete all files in the current directory except for the most recent export """
        files = self.nlst()
        latest = self.latest_filename
        for filename in files:
            if filename != latest:
                result = self.delete(filename)
                logger.info(f"Deleted old export from FTP: {result}")


class Sftp(ImplicitFTP_TLS, Ftp):
    def __init__(
        self,
        url: str,
        username: str,
        password: str,
        destination: str = None,
        basepath: str = None,
        **kwargs,
    ):
        super().__init__()
        self.url = url
        self.username = username
        self.password = password
        self.destination = destination
        self.ssl_version = ssl.PROTOCOL_SSLv23
        self.connect(host=url, port=990)
        self.login(username, password)
        self.m = None  # model hook
        self.basepath = basepath or "."
        self.cwd(self.basepath)


if __name__ == "__main__":
    logging.basicConfig(level=10)
    f = Ftp.from_config()
    result = f.get_latest()

    f.cleanup()

    # fname = result.get("filename")
    # f.delete(fname)



# type: ignore

from __future__ import annotations

import logging
from enum import Enum
from timeit import default_timer as timer
from typing import Dict, List

from sqlalchemy.dialects.postgresql.dml import Insert
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql import func

import metrics
import util
from util.deco import classproperty
from fracx import db


class Operation(Enum):
    INSERT = "insert"
    UPDATE = "update"
    DELETE = "delete"
    MERGE = "merge"


class TimestampMixin(object):
    created_at = db.Column(
        db.DateTime(timezone=True), default=func.now(), nullable=False
    )
    updated_at = db.Column(
        db.DateTime(timezone=True), default=func.now(), nullable=False
    )


logger = logging.getLogger(__name__)


class CoreMixin(object):
    """Base class for sqlalchemy ORM tables containing mostly utility functions for
       accessing table properties and managing insert, update, and upsert operations.
    """

    pks = None

    @classproperty
    def s(self):
        return db.session

    @classproperty
    def pks(cls) -> List[tuple]:  # noqa
        query = cls.s.query().with_entities(*cls.primary_key_columns())
        return query.all()

    @classmethod
    def primary_key_columns(cls) -> List:
        """Returns a list of sqlalchemy column objects for this table's primary keys.

        Returns:
            list -- [Column1, Column2, ...]
        """

        # return [v for k, v in cls.__table__.c.items() if v.primary_key]
        return list(cls.__table__.primary_key.columns)

    @classmethod
    def primary_key_names(cls) -> List[str]:
        """Returns the column names of this table's primary keys.

        Returns:
            list -- column names
        """

        return list(cls.__table__.primary_key.columns.keys())

    @classmethod
    def persist_objects(cls, objects: List[db.Model]):
        cls.s.add_all(objects)
        cls.persist()
        logger.info(
            f"{cls.__table__.name}.persist_objects: inserted {len(objects)} records"
        )

    @classmethod
    def persist(cls) -> None:
        """Propagate changes in session to database.
        """
        try:
            cls.s.flush()
            cls.s.commit()
        except Exception as e:
            logger.info(e)
            cls.s.rollback()

    @classmethod
    def core_insert(
        cls,
        records: List[Dict],
        size: int = None,
        exclude_cols: list = None,
        update_on_conflict: bool = True,
        ignore_on_conflict: bool = False,
    ):
        op_name = "core_insert"
        affected: int = 0
        size = size or len(records)
        exclude_cols = exclude_cols or []
        for chunk in util.chunks(records, size):
            ts = timer()
            chunk = list(chunk)
            stmt = Insert(cls).values(chunk)

            # update these columns when a conflict is encountered
            if ignore_on_conflict:
                final_stmt = stmt.on_conflict_do_nothing(
                    constraint=cls.__table__.primary_key
                )
                op_name = op_name + "_ignore_on_conflict"
            elif update_on_conflict:
                on_conflict_update_cols = [
                    c.name
                    for c in cls.__table__.c
                    if c not in list(cls.__table__.primary_key.columns)
                    and c.name not in exclude_cols
                ]
                op_name = op_name + "_update_on_conflict"

                # append 'on conflict' clause to insert statement
                final_stmt = stmt.on_conflict_do_update(
                    constraint=cls.__table__.primary_key,
                    set_={
                        k: getattr(stmt.excluded, k) for k in on_conflict_update_cols
                    },
                )

            else:
                final_stmt = stmt
            try:
                cls.s.bind.engine.execute(final_stmt)
                cls.persist()
                exc_time = round(timer() - ts, 2)
                n = len(chunk)
                cls.post_op_metrics(Operation.INSERT, op_name, n, exc_time)
                affected += len(records)

            except IntegrityError as ie:
                logger.debug(ie.args[0])

                # fragment and reprocess
                if len(records) > 1:
                    first_half = records[: len(records) // 2]
                    second_half = records[len(records) // 2 :]
                    cls.core_insert(
                        records=first_half,
                        size=len(first_half) // 4,
                        update_on_conflict=update_on_conflict,
                        ignore_on_conflict=ignore_on_conflict,
                    )
                    cls.core_insert(
                        records=second_half,
                        size=len(second_half) // 4,
                        update_on_conflict=update_on_conflict,
                        ignore_on_conflict=ignore_on_conflict,
                    )
            except Exception as e:
                logger.error(e)

        return affected

    @classmethod
    def bulk_insert(cls, records: List[Dict], size: int = None):

        affected: int = 0
        size = size or len(records)

        for chunk in util.chunks(records, size):

            cls.s.bulk_insert_mappings(cls, chunk)
            cls.persist()
            logger.info(
                f"{cls.__table__.name}.bulk_insert: inserted {len(records)} records"
            )
            affected += len(records)
        return affected

    @classmethod
    def bulk_update(cls, records: List[Dict], size: int = None):

        affected: int = 0
        size = size or len(records)

        for chunk in util.chunks(records, size):
            cls.s.bulk_update_mappings(cls, chunk)
            cls.persist()
            logger.info(
                f"{cls.__table__.name}.bulk_update: updated {len(records)} records"
            )
            affected += len(records)
        return affected

    @classmethod
    def bulk_merge(cls, records: List[Dict], size: int = None):
        affected: int = 0
        size = size or len(records)

        for chunk in util.chunks(records, size):
            ts = timer()
            chunk = list(chunk)
            cls.s.add_all([cls.s.merge(cls(**row)) for row in chunk])
            cls.persist()
            exc_time = round(timer() - ts, 2)
            n = len(chunk)
            cls.post_op_metrics(Operation.MERGE, "bulk_merge", n, exc_time)
            affected += len(records)

        return affected

    @classmethod
    def primary_key_values(cls, as_list: bool = False) -> List[tuple]:
        query = cls.s.query().with_entities(*cls.primary_key_columns())
        return query.all()

    @classmethod
    def post_op_metrics(
        cls, method_type: Operation, method: str, n: int, exc_time: float
    ):
        op_name = method_type.name.lower()
        tags = {"tablename": cls.__table__.name, "method": method}
        measurements = {
            f"{op_name}s": n,
            f"{op_name}_time": exc_time,
            f"{op_name}s_per_second": n / exc_time or 1,
        }

        for key, value in measurements.items():
            metrics.post(key, value, tags=tags)

        logger.info(
            f"{cls.__table__.name}.{method}: {op_name}ed {n} records ({exc_time}s)",
            extra=measurements,
        )



from __future__ import annotations
from typing import Dict, List, Union, Iterable
import logging
from pathlib import Path

from flask_sqlalchemy import Model


from api.models import *  # noqa
from collector.endpoint import Endpoint
from collector.transformer import Transformer
from config import get_active_config


logger = logging.getLogger(__name__)

conf = get_active_config()


class Collector(object):
    """ Conduit for transferring newly collected data into a backend data model """

    _tf = None
    _functions = None
    _model = None

    def __init__(
        self,
        endpoint: Endpoint,
        functions: Dict[Union[str, None], Union[str, None]] = None,
    ):
        self.endpoint = endpoint
        self._functions = functions

    @property
    def functions(self):
        if self._functions is None:
            self._functions = conf.functions
        return self._functions

    @property
    def model(self) -> Model:
        if self._model is None:
            self._model = self.endpoint.model
        return self._model

    @property
    def tf(self):
        if self._tf is None:
            self._tf = Transformer(
                aliases=self.endpoint.mappings.get("aliases", {}),
                exclude=self.endpoint.exclude,
                ignore_unknown=self.endpoint.ignore_unknown,
            )
        return self._tf

    def transform(self, data: dict) -> dict:
        return self.tf.transform(data)


class FracScheduleCollector(Collector):
    def collect(
        self,
        iterable: Iterable,
        update_on_conflict: bool = True,
        ignore_on_conflict: bool = False,
    ):

        rows = []
        for row in iterable:
            transformed = self.filter(self.transform(row))
            if transformed:
                rows.append(transformed)

            if len(rows) > 0 and len(rows) % conf.COLLECTOR_WRITE_SIZE == 0:
                self.model.core_insert(
                    rows,
                    update_on_conflict=update_on_conflict,
                    ignore_on_conflict=ignore_on_conflict,
                )
                rows = []

        # persist leftovers
        self.model.core_insert(rows)

    def filter(self, row: Dict) -> Union[Dict, None]:
        if row.get("shllat") and row.get("shllon"):
            return row
        else:
            return None


if __name__ == "__main__":
    from fracx import create_app, db
    from collector import Endpoint
    from collector.filehandler import BytesFileHandler

    logging.basicConfig(level=10)
    logger.setLevel(10)

    app = create_app()
    app.app_context().push()

    endpoints = Endpoint.load_from_config(conf)

    endpoint = endpoints["frac_schedules"]
    c = FracScheduleCollector(endpoint)

    content = b""
    with open("data/bytes.txt", "rb") as f:
        content = f.read()

    iterable = BytesFileHandler.xlsx(
        content, date_columns=endpoint.mappings.get("dates")
    )

    iterable = list(iterable)

    c.collect(iterable)




from __future__ import annotations
from typing import List, Dict

import logging
from config import get_active_config

import util

from collector.parser import Parser

conf = get_active_config()

sp = util.StringProcessor()

logger = logging.getLogger(__name__)


class RowParser(object):
    """ Transform an XML response into a normalized Python object"""

    def __init__(
        self,
        aliases: Dict[str, str] = None,
        exclude: List[str] = None,
        normalize: bool = False,
        parsers: List[Parser] = None,
    ):
        self.normalize = normalize
        self.aliases = aliases or {}
        self.exclude = exclude or []
        self.parsers = parsers or []
        self.errors: List[str] = []

    def __repr__(self):
        s = "s" if len(self.parsers) > 1 else ""
        return f"RowParser: {len(self.parsers)} attached parser{s}"

    def add_parser(
        self, parser: Parser = None, ruleset: Dict[str, List] = None, name: str = None
    ):
        self.parsers.append(parser or Parser.init(ruleset, name=name))
        return self

    def normalize_keys(self, data: Dict) -> Dict:
        return util.apply_transformation(data, sp.normalize, keys=True, values=False)

    def parse_value_dtypes(self, data: Dict) -> Dict:
        for parser in self.parsers:
            data = util.apply_transformation(
                data, parser.parse, keys=False, values=True
            )
        return data

    def parse(self, row: dict, parse_dtypes: bool = True, **kwargs) -> Dict:
        # parsed = self.normalize_keys(row)
        if parse_dtypes:
            parsed = self.parse_value_dtypes(row)
        return parsed

    @staticmethod
    def load_from_config(parser_conf: dict):
        parser_conf = parser_conf.get("parsers", parser_conf)
        parsers: List[Parser] = []
        for name, parser_def in parser_conf.items():
            ruleset = parser_def.get("rules", parser_def)
            parsers.append(Parser.init(ruleset, name))

        return RowParser(parsers=parsers)


if __name__ == "__main__":
    rp = RowParser.load_from_config(conf.PARSER_CONFIG)

    row = {
        "region": "PMI",
        "company": "Example",
        "well_name": "Example 1-30H",
        "well_api": "42461405550000",
        "frac_start_date": 43798.74804875,
        "frac_end_date": 43838.74804875,
        "surface_lat": "32.4150535",
        "surface_long": "-101.6295689",
        "bottomhole_lat": "",
        "bottomhole_long": "",
        "tvd": "8323",
        "target_formation": "Wolfcamp B",
        "start_in": "STARTED",
        "duration_in_days": "40",
        "reviewed_by": "",
        "risk_assessments": "",
        "comments": "",
        "last_upload_date": "2020-01-03 17:57:12.505",
    }
    rp.parse(row)



import logging

from sqlalchemy.sql import func

from api.mixins import CoreMixin
from config import get_active_config
from fracx import db

conf = get_active_config()

logger = logging.getLogger(__name__)


class FracSchedule(CoreMixin, db.Model):

    __tablename__ = conf.FRAC_SCHEDULE_TABLE_NAME

    api14 = db.Column(db.String(14), nullable=False, primary_key=True)
    api10 = db.Column(db.String(10), nullable=False)
    wellname = db.Column(db.String(), nullable=True)
    operator = db.Column(db.String())
    frac_start_date = db.Column(db.Date(), primary_key=True)
    frac_end_date = db.Column(db.Date(), primary_key=True)
    status = db.Column(db.String())
    tvd = db.Column(db.Integer())
    target_formation = db.Column(db.String())
    shllat = db.Column(db.Float())
    shllon = db.Column(db.Float())
    bhllat = db.Column(db.Float())
    bhllon = db.Column(db.Float())
    created_at = db.Column(
        db.DateTime(timezone=True), default=func.now(), nullable=False
    )
    updated_at = db.Column(
        db.DateTime(timezone=True), default=func.now(), nullable=False
    )



import logging
import os
import shutil
import sys
from collections import defaultdict
import subprocess

import click
from flask.cli import FlaskGroup, AppGroup

from fracx import create_app, db
from config import get_active_config

logger = logging.getLogger()


CONTEXT_SETTINGS = dict(
    help_option_names=["-h", "--help"], ignore_unknown_options=False
)

conf = get_active_config()
app = create_app()
cli = FlaskGroup(create_app=create_app, context_settings=CONTEXT_SETTINGS)
run_cli = AppGroup("run")
test_cli = AppGroup("test")


def get_terminal_columns():
    return shutil.get_terminal_size().columns


def hr():
    return "-" * get_terminal_columns()


@cli.command()
def ipython_embed():
    """Runs a ipython shell in the app context."""
    try:
        import IPython
    except ImportError:
        click.echo("IPython not found. Install with: 'pip install ipython'")
        return
    from flask.globals import _app_ctx_stack

    app = _app_ctx_stack.top.app
    banner = "Python %s on %s\nIPython: %s\nApp: %s%s\nInstance: %s\n" % (
        sys.version,
        sys.platform,
        IPython.__version__,
        app.import_name,
        app.debug and " [debug]" or "",
        app.instance_path,
    )

    ctx = {}

    # Support the regular Python interpreter startup script if someone
    # is using it.
    startup = os.environ.get("PYTHONSTARTUP")
    if startup and os.path.isfile(startup):
        with open(startup, "r") as f:
            eval(compile(f.read(), startup, "exec"), ctx)

    ctx.update(app.make_shell_context())

    IPython.embed(banner1=banner, user_ns=ctx)


@run_cli.command()
@click.option(
    "update_on_conflict",
    "--update-on-conflict",
    "-u",
    help="Prevent updating records that already exist",
    show_default=True,
    default=True,
)
@click.option(
    "ignore_on_conflict",
    "--ignore-conflict",
    "-i",
    help="Ignore records that already exist",
    show_default=True,
    is_flag=True,
)
@click.option(
    "use_existing",
    "--use-existing",
    "-e",
    help=f"Use a previously downloaded file",
    is_flag=True,
)
def collector(update_on_conflict, ignore_on_conflict, use_existing):
    "Run a one-off task to synchronize from the fracx data source"
    from collector import Endpoint, FracScheduleCollector, Ftp, BytesFileHandler

    logger.info(conf)

    endpoint = Endpoint.load_from_config(conf)["frac_schedules"]
    collector = FracScheduleCollector(endpoint)

    ftp = Ftp.from_config()
    latest = ftp.get_latest()

    rows = BytesFileHandler.xlsx(
        latest.get("content"), date_columns=endpoint.mappings.get("dates")
    )
    collector.collect(rows, update_on_conflict, ignore_on_conflict)

    ftp.cleanup()


@cli.command()
def endpoints():
    from collector import Endpoint

    for name, ep in Endpoint.load_from_config(conf).items():
        click.secho(name)


def main(argv=sys.argv):
    """
    Args:
        argv (list): List of arguments
    Returns:
        int: A return code
    Does stuff.
    """

    cli()
    return 0


cli.add_command(run_cli)

if __name__ == "__main__":
    cli()



from typing import Generator, Dict, List, Union
from datetime import datetime

import xlrd
from util import StringProcessor

sp = StringProcessor()


class BytesFileHandler:
    @classmethod
    def xlsx(
        cls, content: bytes, sheet_no: int = 0, date_columns: List[str] = None
    ) -> Generator[Dict, None, None]:
        """ Extract the data of an Excel sheet from a byte stream """
        date_columns = date_columns or []
        sheet = xlrd.open_workbook(file_contents=content).sheet_by_index(sheet_no)

        keys = sheet.row_values(0)
        keys = [sp.normalize(x) for x in keys]

        for idx in range(1, sheet.nrows):
            result = dict(zip(keys, sheet.row_values(idx)))
            for dc in date_columns:
                value = result.get(dc)
                # print(f"{dc=}, {value=}")
                result[dc] = cls._parse_excel_date(value, sheet.book.datemode)

            yield result

    @classmethod
    def _parse_excel_date(cls, value: Union[float, None], date_mode: int = 0):
        if value:
            return datetime(*xlrd.xldate_as_tuple(value, date_mode))
        else:
            return value


if __name__ == "__main__":
    content = b""
    with open("data/bytes.txt", "rb") as f:
        content = f.read()

    rows = BytesFileHandler.xlsx(
        content, date_columns=["frac_start_date", "frac_end_date"]
    )

    # rows = BytesFileHandler.xlsx(content)

    for row in rows:
        print(row)
        break

        cls = BytesFileHandler
        sheet_no = 0
        BytesFileHandler._parse_excel_date(value, date_mode)

    datetime(*xlrd.xldate_as_tuple(43798.74804875, 0))



from typing import Callable, Union, Iterable, Generator, Dict
import math
import json
import urllib.parse
import itertools

from util.strings import StringProcessor  # noqa
from util.jsontools import DateTimeEncoder


def hf_size(size_bytes: Union[str, int]) -> str:
    """Human friendly string representation of a size in bytes.

    Source: https://stackoverflow.com/questions/5194057/better-way-to-convert-file-sizes-in-python

    Arguments:
        size_bytes {Union[str, int]} -- size of object in number of bytes

    Returns:
        str -- string representation of object size. Ex: 299553704 -> "285.68 MB"
    """
    if size_bytes == 0:
        return "0B"

    suffixes = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")

    if isinstance(size_bytes, str):
        size_bytes = int(size_bytes)

    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return f"{s} {suffixes[i]}"


def chunks(iterable: Iterable, n: int = 1000) -> Generator:
    """ Process an interable in chunks of size n (default=1000) """
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, n)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)


def apply_transformation(
    data: dict, convert: Callable, keys: bool = False, values: bool = True
) -> Dict:
    """ Recursively apply the passed function to a dict's keys, values, or both """
    if isinstance(data, (str, int, float)):
        if values:
            return convert(data)
        else:
            return data
    if isinstance(data, dict):
        new = data.__class__()
        for k, v in data.items():
            if keys:
                new[convert(k)] = apply_transformation(v, convert, keys, values)
            else:
                new[k] = apply_transformation(v, convert, keys, values)
    elif isinstance(data, (list, set, tuple)):
        new = data.__class__(
            apply_transformation(v, convert, keys, values) for v in data
        )
    else:
        return data
    return new


def urljoin(base: str = "", path: str = "") -> str:
    base = base or ""
    path = path or ""
    if not base.endswith("/"):
        base = base + "/"
    if path.startswith("/"):
        path = path[1:]
    return urllib.parse.urljoin(base, path)


def to_json(d: dict, path: str, cls=DateTimeEncoder):
    with open(path, "w") as f:
        json.dump(d, f, cls=cls, indent=4)


def load_json(path: str) -> Dict:
    with open(path, "r") as f:
        return json.load(f)


def query_dict(path: str, data: dict, sep: str = ".") -> Union[Dict, None]:
    elements = path.split(sep)
    for e in elements:
        if not issubclass(type(data), dict):
            raise ValueError(f"{data} ({type(data)}) is not a subclass of dict")
        data = data.get(e, {})
    return data if data != {} else None



# pylint: disable=arguments-differ, method-hidden

from typing import Any
import json
from datetime import datetime, date, timedelta


class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        return super(DateTimeEncoder, self).default(obj)


class ObjectEncoder(json.JSONEncoder):
    """Class to convert an object into JSON."""

    def default(self, obj: Any):
        """Convert `obj` to JSON."""
        if hasattr(obj, "to_json"):
            return self.default(obj.to_json())
        elif hasattr(obj, "__dict__"):
            return obj.__class__.__name__
        elif hasattr(obj, "tb_frame"):
            return "traceback"
        elif isinstance(obj, timedelta):
            return obj.__str__()
        else:
            # generic, captures all python classes irrespective.
            cls = type(obj)
            result = {
                "__custom__": True,
                "__module__": cls.__module__,
                "__name__": cls.__name__,
            }
            return result



from typing import List, Tuple, Union, Dict
import logging


from config import get_active_config, project

logger = logging.getLogger(__name__)
conf = get_active_config()

datadog = None


def load():
    """ Load and initialize the Datadog library """
    try:
        parms = conf.datadog_params
        if parms.get("enabled"):
            logger.debug("Datadog Enabled")
            if parms.get("api_key") and parms.get("app_key"):
                global datadog  # pylint: disable=global-statement
                import datadog

                datadog.initialize(**parms)
                logger.info("Datadog initialized")
            else:
                missing_key = "api_key" if not parms.get("api_key") else "app_key"
                logger.error(
                    f"Failed to load Datadog configuration: missing {missing_key}"
                )
        else:
            logger.debug("Datadog disabled.")

    except Exception as e:
        logger.error(f"Failed to load Datadog configuration: {e}")


def post(
    name: str,
    points: Union[int, float, List[Tuple]],
    metric_type: str = "count",
    tags: Union[Dict, List, str] = None,
):
    """ Send a metric through the Datadog http api.

        Example:
                    api.Metric.post(
                        metric='my.series',
                        points=[
                            (now, 15),
                            (future_10s, 16)
                        ],
                        metric_type="count",
                        tags=["tag1", "tag2"]
                    )

    Arguments:
        name {str} -- metric name
        points {Union[int, float, List[Tuple]]} -- metric value(s)
    """
    try:
        name = f"{project}.{name}".lower()
        if datadog:
            result = datadog.api.Metric.send(
                metric=name,
                points=points,
                type=str(metric_type).lower(),
                tags=to_tags(conf.DEFAULT_TAGS) + to_tags(tags or []),
            )
            if result.get("status") == "ok":
                logger.debug(
                    "Datadog metric successfully sent: name=%s, points=%s",
                    name,
                    points,
                )
            else:
                logger.debug(
                    "Problem sending Datadog metric: status=%s, name=%s, points=%s",
                    result.get("status"),
                    name,
                    points,
                )
        else:
            logger.debug(
                "Datadog not configured. Suppressing metric name=%s, points=%s",
                name,
                points,
            )
    except Exception as e:
        logger.debug("Failed to send Datadog metric: %s", e)


def post_event(title: str, text: str, tags: Union[Dict, List, str] = None):
    """ Send an event through the Datadog http api. """
    try:
        if datadog:
            datadog.api.Event.create(title=title, text=text, tags=to_tags(tags or []))
    except Exception as e:
        logger.debug("Failed to send Datadog event: %s", e)


def post_heartbeat():
    """ Send service heartbeat to Datadog """
    return post("heartbeat", 1, metric_type="gauge")


def to_tags(values: Union[Dict, List, str], sep: str = ",") -> List[str]:
    """ Coerce the passed values into a list of colon separated key-value pairs.

        dict example:
            {"tag1": "value1", "tag2": "value2", ...} -> ["tag1:value1", "tag2:value2", ...]

        list example:
            ["tag1", "tag2", ...] -> ["tag1", "tag2", ...]

        str example (comma-delimited):
            "tag1:value1, tag2:value2", ..." -> ["tag1:value1", "tag2:value2", ...]

        str example (single):
            "tag1:value1" -> ["tag1:value1"]
    """
    result: List[str] = []
    if isinstance(values, dict):
        result = [
            f"{key}:{str(value).lower().replace(' ','_')}"
            for key, value in values.items()
            if isinstance(value, (str, int))
        ]
    elif isinstance(values, str):
        if "," in values:
            result = values.split(sep)
        else:
            result = [values]
    elif isinstance(values, list):
        result = values
    else:
        result = []

    return result



from __future__ import annotations
from typing import Any, Callable, List, Union, Dict
import functools
import logging
import re
from datetime import datetime
import dateutil.parser

from config import get_active_config

conf = get_active_config()

logger = logging.getLogger(__name__)


def safe_convert(func):
    """ Generic error handling decorator for primative type casts """

    @functools.wraps(func)
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger.debug(f"{func} failed: {e}")
            return None

    return func_wrapper


def locate_resource(name: str) -> Callable:
    """ Locate a resource (module, func, etc) within the module's namespace """
    resource = globals()[name]
    return resource


class Criterion:
    """ Basic component of validation logic used to compose a parsing rule """

    def __init__(self, func: Callable, name: str = None):
        self.name = name or ""
        self.func = func

    def __call__(self, value: Any):
        return bool(self.func(value))

    def __repr__(self):
        return f"Criterion: {self.name} - {self.func}"


class RegexCriterion(Criterion):
    """ Regex extraction harness for parser rule"""

    def __init__(self, regex: str, name: str = None):
        self.pattern = regex
        self.regex = re.compile(regex)
        super().__init__(func=self.regex.match, name=name)

    def __call__(self, value: Any):
        return super().__call__(str(value))

    def __repr__(self):
        return f"RegexCriterion: {self.name} - {self.pattern}"


class TypeCriterion(Criterion):
    """ Type check harness for parser rule """

    def __init__(self, dtype: type, name: str = None):
        func = lambda v: isinstance(v, dtype)  # noqa
        super().__init__(func=func, name=name)


class ValueCriterion(Criterion):
    """ Value comparison harness for parser rule """

    def __init__(self, value: Union[str, int, float, bool], name: str = None):
        func = lambda v: v == value  # noqa
        super().__init__(func=func, name=name)


class ParserRule:
    """ Validation rule used by a Parser to determine if/how to parse a value """

    def __init__(
        self,
        criteria: List[Criterion],
        name: str = None,
        allow_partial: bool = True,
        **kwargs,
    ):
        """
        Arguments:
            criteria {List[Criterion]} -- Criteria list

        Keyword Arguments:
            name {str} -- Rule name
            partial {bool} -- If True, the rule will pass if any criteria are satisfied.
            If False, the rule will pass only if all criteria are satisfied.

        """
        self.name = name or ""
        self.criteria = criteria
        self.allow_partial = allow_partial

    def __repr__(self):
        size = len(self.criteria)
        return f"ParserRule:{self.name} ({self.match_mode}) -  {size} criteria"

    def __call__(
        self, value: Any, return_partials: bool = False
    ) -> Union[bool, List[bool]]:
        """ Enables the ParserRule to be callable, such that invoking the rule with
            a passed value will return the evaluation result of the called rule.

            Example: MyIntegerParserRule("13") -> True

            return_partials: set to True to return a list of the result of each
                             criteria. If set to False (default), return the result of
                             the appropriate boolean operation:

                            if self.allow_partials = True: ->
                                return any([partial1, partial2, ...]
                            else:
                                return all([partial1, partial2, ...]
        """
        partials = [c(value) for c in self.criteria]
        if return_partials:
            return partials
        if self.allow_partial:
            return any(partials)
        else:
            return all(partials)

    @property
    def match_mode(self):
        """ Indicates if all criteria must be met to consider a parse successful """
        return "PARTIAL" if self.allow_partial else "FULL"

    @classmethod
    def from_list(cls, criteria: List[Dict], **kwargs) -> ParserRule:
        """ Initialize a rule from a list of criteria specifications.
                Example criteria spec:
                    criteria = \
                        [
                            {
                                "name": "parse_integers",
                                "type": "RegexCriterion",
                                "value": r"^[-+]?[0-9]+$",
                            },
                        ],
         """
        criteriaObjs: List[Criterion] = []
        for c in criteria:
            CriteriaType = locate_resource(c["type"])
            criteriaObjs.append(CriteriaType(c["value"], c["name"]))
        return cls(criteriaObjs, **kwargs)


class Parser:
    """ Parses text values according to a set of arbitrary rules """

    def __init__(
        self, rules: List[ParserRule], name: str = None, parse_dtypes: bool = True
    ):
        self.name = name or ""
        self.rules = rules
        self.parse_dtypes = parse_dtypes

    def __repr__(self):
        return f"Parser - {self.name}: {len(self.rules)} rules"

    @classmethod
    def init(cls, ruleset: Dict[str, List], name: str = None) -> Parser:
        """ Initialize from a configuration dict """
        rules: List[ParserRule] = []
        for ruledef in ruleset:
            rules.append(ParserRule.from_list(**ruledef))  # type: ignore
        return cls(rules, name=name)

    @staticmethod
    @safe_convert
    def try_int(s: str) -> int:
        # if str(s).replace("+", "").replace("-", "").isdigit():
        # return int(float(s))
        return int(s)

    @staticmethod
    @safe_convert
    def try_float(s: str) -> float:
        return float(s)

    @staticmethod
    @safe_convert
    def try_date(s: str) -> datetime:
        if s is not None:
            return dateutil.parser.parse(s)
        else:
            return s

    @staticmethod
    @safe_convert
    def try_bool(s: str) -> Union[bool, str]:
        value = str(s)
        if value.lower() == "true":
            return True
        elif value.lower() == "false":
            return False
        else:
            return s

    @staticmethod
    @safe_convert
    def try_empty_str_to_none(s: str) -> Union[None, str]:
        return None if s == "" else s

    def add_rule(self, rule: ParserRule):
        self.rules.append(rule)

    def run_checks(
        self, value: Any, return_partials: bool = False
    ) -> Union[bool, List[bool]]:
        """ Check if all parsing rules are satisfied """
        checks = []
        for Rule in self.rules:
            result = Rule(value)
            checks.append(result)
            if not result:
                logger.debug("Parser check failed: %s", (Rule,))
            else:
                logger.debug("Parser check passed: %s", (Rule,))

        return all(checks) if not return_partials else checks  # type: ignore

    def parse_dtype(self, value: str) -> Union[int, float, str, datetime]:
        funcs = [
            "try_int",
            "try_float",
            "try_date",
            "try_bool",
        ]

        for fname in funcs:
            func = getattr(self, fname)
            newvalue = func(value)
            logger.debug(
                f"Parsed dtype: %s -> %s (%s)",
                value or "None",
                newvalue or "None",
                type(newvalue).__name__,
            )
            if not isinstance(newvalue, str) and newvalue is not None:
                value = newvalue
                break

        value = self.try_empty_str_to_none(value)
        return value

    def parse(self, value: Any) -> Any:
        """ Attempt to parse a value if all checks are satisfied """
        if not self.run_checks(value):
            return value
        else:
            return self.parse_dtype(value) if self.parse_dtypes else value


if __name__ == "__main__":

    parser = Parser.init(
        conf.PARSER_CONFIG["parsers"]["default"]["rules"], name="default"
    )

    test_values = [
        None,
        "2019-01-01",
        "1",
        "+1",
        "-1",
        "2018",
        "",
        "0",
        "+0",
        "-0",
        "11",
        "00",
        "01",
        "+11",
        "+00",
        "+01",
        "-11",
        "-00",
        "-01",
        "1234567890",
        "+1234567890",
        "-1234567890",
        "1.1034",
        "+1.1034",
        "-1.1034",
        "0.1034",
        "+0.1034",
        "-0.1034",
        "11.1034",
        "00.1034",
        "01.1034",
        "+11.1034",
        "+00.1034",
        "+01.1034",
        "-11.1034",
        "-00.1034",
        "-01.1034",
        "1234567890.1034",
        "+1234567890.1034",
        "-1234567890.1034",
        "2019-01-01",
        "2020-06-06",
        "19-01-01",
        "2019-01",
        "qwe2019-01-01",
        "2019-01-01rte",
        "3242019-01-01",
        "31.24141",
        "101.98853",
        "+31.24141",
        "-101.98853",
        "9/11/2014 12:00:00 AM",
        "9/25/2014 5:00:00 AM",
        "true",
        "false",
        "True",
        "False",
        "42461405550000",
    ]

    logger.setLevel(20)
    results = []
    for value in test_values:
        new_value = parser.parse(value)
        new_value = new_value if new_value is not None else "-"
        value = value if value is not None else "-"
        results.append(new_value)
        print(f"{value:<20} -> {new_value} ({type(new_value).__name__})")



from __future__ import annotations
from typing import List
import logging
import os
import socket
import shutil

import tomlkit
import yaml
from attrdict import AttrDict
from sqlalchemy.engine.url import URL

_pg_aliases = ["postgres", "postgresql", "psycopg2", "psycopg2-binary"]
_mssql_aliases = ["mssql", "sql server"]

APP_SETTINGS = os.getenv("APP_SETTINGS", "fracx.config.ProductionConfig")
FLASK_APP = os.getenv("FLASK_APP", "fracx.manage.py")


def abs_path(path: str, filename: str) -> str:
    return os.path.abspath(os.path.join(path, filename))


def make_config_path(path: str, filename: str) -> str:
    return os.path.abspath(os.path.join(path, filename))


def load_config(path: str) -> AttrDict:
    try:
        with open(path) as f:
            return AttrDict(yaml.safe_load(f))
    except FileNotFoundError as fe:
        print(f"Failed to load configuration: {fe}")


def get_active_config() -> AttrDict:
    return globals()[APP_SETTINGS.replace("fracx.config.", "")]()


def get_default_port(driver: str):
    port = None
    if driver in _pg_aliases:
        port = 5432
    elif driver in _mssql_aliases:
        port = 1433

    return port


def get_default_driver(dialect: str):
    driver = None
    if dialect in _pg_aliases:
        driver = "postgres"  # "psycopg2"
    elif dialect in _mssql_aliases:
        driver = "pymssql"

    return driver


def get_default_schema(dialect: str):
    driver = None
    if dialect in _pg_aliases:
        driver = "public"
    elif dialect in _mssql_aliases:
        driver = "dbo"

    return driver


def make_url(url_params: dict) -> URL:
    return URL(**url_params)


def split_args(args: List[str]) -> List:
    return [x for x in args]


def _get_project_meta() -> dict:
    pyproj_path = "./pyproject.toml"
    if os.path.exists(pyproj_path):
        with open(pyproj_path, "r") as pyproject:
            file_contents = pyproject.read()
        return tomlkit.parse(file_contents)["tool"]["poetry"]
    else:
        return {}


pkg_meta = _get_project_meta()
project = pkg_meta.get("name")
version = pkg_meta.get("version")


class BaseConfig:
    """Base configuration"""

    FLASK_ENV = os.getenv("FLASK_ENV", "development")
    ENV_NAME = os.getenv("ENV_NAME", socket.gethostname())

    """ Datadog """
    DATADOG_ENABLED = os.getenv("DATADOG_ENABLED", False)
    DATADOG_API_KEY = os.getenv("DATADOG_API_KEY", None)
    DATADOG_APP_KEY = os.getenv("DATADOG_APP_KEY", None)

    """ General """
    CONFIG_BASEPATH = "./config"

    """ Collector """
    COLLECTOR_CONFIG_PATH = make_config_path(CONFIG_BASEPATH, "collector.yaml")
    COLLECTOR_CONFIG = load_config(COLLECTOR_CONFIG_PATH)
    COLLECTOR_FTP_URL = os.getenv("FRACX_FTP_URL", "sftp.pdswdx.com")
    COLLECTOR_FTP_PORT = os.getenv("FRACX_FTP_URL", "21")
    COLLECTOR_FTP_OUTPATH = os.getenv("FRACX_FTP_OUTPATH", "/Outbound")
    COLLECTOR_FTP_INPATH = os.getenv("FRACX_FTP_INPATH", "/Inbound")
    COLLECTOR_FTP_USERNAME = os.getenv("FRACX_FTP_USERNAME")
    COLLECTOR_FTP_PASSWORD = os.getenv("FRACX_FTP_PASSWORD")
    COLLECTOR_FILE_PREFIX = os.getenv("FRACX_FILE_PREFIX")
    COLLECTOR_WRITE_SIZE = int(os.getenv("COLLECTOR_WRITE_SIZE", "1000"))

    """ Parser """
    PARSER_CONFIG_PATH = abs_path(CONFIG_BASEPATH, "parsers.yaml")
    PARSER_CONFIG = load_config(PARSER_CONFIG_PATH)

    """ Logging """
    LOG_LEVEL = os.getenv("LOG_LEVEL", logging.INFO)
    LOG_FORMAT = os.getenv("LOG_FORMAT", "layman")

    """ --------------- Sqlalchemy --------------- """
    DATABASE_DIALECT = os.getenv("DATABASE_DIALECT", "postgres")
    DATABASE_DRIVER = os.getenv("DATABASE_DRIVER", get_default_driver(DATABASE_DIALECT))
    DATABASE_USERNAME = os.getenv("DATABASE_USERNAME", "")
    DATABASE_PASSWORD = os.getenv("DATABASE_PASSWORD", "")
    DATABASE_HOST = os.getenv("DATABASE_HOST", "localhost")
    DATABASE_PORT = os.getenv("DATABASE_PORT", get_default_port(DATABASE_DRIVER))
    DATABASE_SCHEMA = os.getenv("DATABASE_SCHEMA", get_default_schema(DATABASE_DRIVER))
    DATABASE_NAME = os.getenv("DATABASE_NAME", "postgres")
    DATABASE_URL_PARAMS = {
        "drivername": DATABASE_DRIVER,
        "username": DATABASE_USERNAME,
        "password": DATABASE_PASSWORD,
        "host": DATABASE_HOST,
        "port": DATABASE_PORT,
        "database": DATABASE_NAME,
    }
    SQLALCHEMY_DATABASE_URI = str(make_url(DATABASE_URL_PARAMS))
    FRAC_SCHEDULE_TABLE_NAME = os.getenv("FRAC_SCHEDULE_TABLE_NAME", "frac_schedules")

    @property
    def show(self):
        return [x for x in dir(self) if not x.startswith("_")]

    @property
    def collector_params(self):
        return self.with_prefix("collector")

    @property
    def datadog_params(self):
        return self.with_prefix("datadog")

    @property
    def endpoints(self):
        return self.COLLECTOR_CONFIG.endpoints

    def __repr__(self):
        """ Print noteworthy configuration items """
        hr = "-" * shutil.get_terminal_size().columns + "\n"
        tpl = "{name:>25} {value:<50}\n"
        string = ""
        string += tpl.format(name="app config:", value=APP_SETTINGS)
        string += tpl.format(name="flask app:", value=FLASK_APP)
        string += tpl.format(name="flask env:", value=self.FLASK_ENV)
        string += tpl.format(
            name="backend:", value=make_url(self.DATABASE_URL_PARAMS).__repr__()
        )
        string += tpl.format(name="collector:", value=self.COLLECTOR_FTP_URL)
        return hr + string + hr

    def with_prefix(self, kw: str):
        """ Return all parameters that begin with the given string.

            Example: kw = "collector"

                Returns:
                    {
                        "base_url": "example.com/api",
                        "path": "path/to/data",
                        "endpoints": {...}
                    }
        """
        if not kw.endswith("_"):
            kw = kw + "_"
        return {
            key.lower().replace(kw.lower(), ""): getattr(self, key)
            for key in dir(self)
            if key.startswith(kw.upper())
        }


class DevelopmentConfig(BaseConfig):
    """Development configuration"""

    DEBUG_TB_ENABLED = True
    SECRET_KEY = os.getenv("SECRET_KEY", "test")
    SQLALCHEMY_TRACK_MODIFICATIONS = True


class TestingConfig(BaseConfig):
    """Testing configuration"""

    CONFIG_BASEPATH = "./tests/data"
    # COLLECTOR_CONFIG_PATH = make_config_path(CONFIG_BASEPATH, "collector.yaml")
    # COLLECTOR_CONFIG = load_config(COLLECTOR_CONFIG_PATH)
    TESTING = True
    TOKEN_EXPIRATION_DAYS = 0
    TOKEN_EXPIRATION_SECONDS = 3
    SQLALCHEMY_TRACK_MODIFICATIONS = False

    API_BASE_URL = "https://api.example.com/v3"
    API_CLIENT_ID = "test_client_id"
    API_CLIENT_SECRET = "test_client_secret"
    API_USERNAME = "username"
    API_PASSWORD = "password"
    API_TOKEN_PATH = "/auth"
    API_DEFAULT_PAGESIZE = 100


class CIConfig(BaseConfig):
    pass


class ProductionConfig(BaseConfig):
    """Production configuration"""

    SQLALCHEMY_TRACK_MODIFICATIONS = False
    CELERYD_PREFETCH_MULTIPLIER = 8
    CELERYD_CONCURRENCY = 12


if __name__ == "__main__":
    c = BaseConfig()



from __future__ import annotations
from typing import Dict, List, Union
import logging
from datetime import date, datetime


from collector.parser import Parser
from collector.row_parser import RowParser
from config import get_active_config


conf = get_active_config()

logger = logging.getLogger(__name__)

Scalar = Union[int, float, str, None, datetime, date]
Row = Dict[str, Scalar]


class TransformationError(Exception):
    pass


class Transformer(object):
    parser = RowParser.load_from_config(conf.PARSER_CONFIG)

    def __init__(
        self,
        aliases: Dict[str, str] = None,
        exclude: List[str] = None,
        normalize: bool = False,
        parser: Parser = None,
        ignore_unknown: bool = True,
    ):
        self.normalize = normalize
        self.aliases = aliases or {}
        self.exclude = exclude or []
        self.errors: List[str] = []
        self.parser = parser or self.parser
        self.ignore_unknown = ignore_unknown

    def __repr__(self):
        la = len(self.aliases)
        le = len(self.exclude)
        unknown = "permissive" if self.ignore_unknown else "strict"
        return f"Transformer: {la} aliases, {le} exclusions ({unknown})"

    def transform(self, row: dict) -> Row:

        try:
            row = self.drop_exclusions(row)
            row = self.apply_aliases(row)
            row = self.parser.parse(row)

            if "api14" in row.keys():
                api14 = str(row["api14"])
                ndiff = 14 - len(api14)
                if ndiff > 0:
                    api14 += "0" * ndiff
                row["api14"] = api14
                row["api10"] = row["api14"][:10]

            numerrs = len(self.errors)
            if len(self.errors) > 0:
                logger.warning(
                    "Captured %s parsing errors during transformation: %s",
                    numerrs,
                    self.errors,
                )

            return row
        except Exception as e:
            logger.exception(f"Transformation error: {e}")
            raise TransformationError(e)

    def apply_aliases(self, row: Row) -> Row:
        return {self.aliases[k]: v for k, v in row.items() if k in self.aliases.keys()}

    def drop_exclusions(self, row: Row) -> Row:
        keys = list(row.keys())
        exclude = [*self.exclude]
        if self.ignore_unknown:
            exclude += [k for k in keys if k not in self.aliases.keys()]
        if len(exclude) > 0:
            logger.debug(f"Dropping {len(exclude)} columns: {exclude}")
            try:
                row = {k: v for k, v in row.items() if k not in exclude}
            except Exception as e:
                msg = f"Failed attempting to drop columns -- {e}"
                self.errors.append(msg)
                logger.debug(msg)
        return row


if __name__ == "__main__":

    from collector import Endpoint

    logging.basicConfig(level=10)
    logger.setLevel(10)

    ep = Endpoint.load_from_config(conf)["frac_schedules"]
    t = Transformer(ep.mappings.aliases, ep.exclude)

    row = {
        "region": "PMI",
        "company": "Example",
        "well_name": "Example 1-30H",
        "well_api": "4246140555",
        "frac_start_date": 43798.74804875,
        "frac_end_date": 43838.74804875,
        "surface_lat": "32.4150535",
        "surface_long": "-101.6295689",
        "bottomhole_lat": "32.4150535",
        "bottomhole_long": "-101.6295689",
        "tvd": "8323",
        "target_formation": "Wolfcamp B",
        "start_in": "STARTED",
        "duration_in_days": "40",
        "reviewed_by": "",
        "risk_assessments": "",
        "comments": "",
        "last_upload_date": "2020-01-03 17:57:12.505",
    }

    parsed = t.transform(row)

    parsed
    from datetime import datetime

    # datetime.fromordinal(parsed["frac_start_date"])

    t.errors



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fracx', 'fracx.collector', 'fracx.metrics', 'fracx.util']

package_data = \
{'': ['*'], 'fracx': ['api/*']}

install_requires = \
['attr>=0.3.1,<0.4.0',
 'attrdict>=2.0.1,<3.0.0',
 'attrs>=19.3.0,<20.0.0',
 'click>=7.0,<8.0',
 'colorama>=0.4.3,<0.5.0',
 'datadog>=0.33.0,<0.34.0',
 'flask-sqlalchemy>=2.4.1,<3.0.0',
 'json_log_formatter>=0.2.0,<0.3.0',
 'logutils>=0.3.5,<0.4.0',
 'psycopg2-binary>=2.8.4,<3.0.0',
 'pyparsing>=2.4.6,<3.0.0',
 'pyproj>=2.4.2,<3.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'python-dotenv>=0.10.3,<0.11.0',
 'pyyaml>=5.2,<6.0',
 'spacy>=2.2.3,<3.0.0',
 'sqlalchemy>=1.3.12,<2.0.0',
 'tomlkit>=0.5.8,<0.6.0',
 'xlrd>=1.2.0,<2.0.0']

entry_points = \
{'console_scripts': ['fracx = fracx.manage:main']}

setup_kwargs = {
    'name': 'fracx',
    'version': '0.1.2',
    'description': "This project is used to interface with [PDSEnergy's](https://pdsenergy.com/frac-interference-exchange/) FracX platform to programmatically submit and download frac schedules.",
    'long_description': '# permian-frac-exchange\n\n<div style="text-align:center;">\n  <table >\n    <tr>\n      <a href="https://codecov.io/gh/la-mar/permian-frac-exchange">\n        <img src="https://codecov.io/gh/la-mar/permian-frac-exchange/branch/master/graph/badge.svg" />\n      </a>\n      <a href="(https://circleci.com/gh/la-mar/permian-frac-exchange">\n        <img src="https://circleci.com/gh/la-mar/permian-frac-exchange.svg?style=svg" />\n      </a>\n    </tr>\n  </table>\n</div>\n\nThis project is used to interface with [PDSEnergy\'s](https://pdsenergy.com/frac-interference-exchange/) FracX platform to programmatically submit and download frac schedules.\n\nCurrently, the project interfaces with the FTP import/export service provided by PDS. As other means of integration become available from PDS, those means will be incorporated into this project as additional ways to import and export data from FracX.\n\n<br/>\n\n## Getting Started\n\n1. Define the necessary environment variables in the docker-compose.yml file or in a .env file in the project root directory (examples below).\n2. Run the container with docker-compose: `docker-compose up`\n   <br/>\n\n## Usage\n\n- Run the container using docker-compose: `docker-compose up`\n\nExample output:\n\n<pre>\n  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n                app config: fracx.config.ProductionConfig\n                flask app: fracx.manage.py\n                flask env: development\n                backend: postgres://fracx:***@db:5432/driftwood\n                collector: sftp.pdswdx.com\n  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\nStarting download from sftp.pdswdx.com...\nDownload successful (download size: 70.58 KB, download_time: 0.0s)\nfrac_schedules.core_insert_update_on_conflict: inserted 625 records (1.28s)\n</pre>\n<br/>\n\n## Configuration\n\nThe application can be configured with environment variables that are passed into the container at runtime. Environment variables can either be defined at the system level or in a file named \'.env\' in the project\'s root directory.\n<br/>\n\n### Saving output to a Postgres database\n\n<br/>\n\n##### Using a .env file\n\nExample configuration with .env file:\n\n.env\n\n```python\n\nFRACX_FTP_USERNAME=YOUR_FRACX_FTP_UERNAME\nFRACX_FTP_PASSWORD=YOUR_FRACX_FTP_PASSWORD\nDATABASE_USERNAME=YOUR_DATABASE_USERNAME\nDATABASE_PASSWORD=YOUR_DATABASE_PASSWORD\nDATABASE_HOST=YOUR_DATABASE_HOST\nDATABASE_NAME=YOUR_DATABASE_NAME\nDATABASE_SCHEMA=fracx\nFRAC_SCHEDULE_TABLE_NAME=frac_schedules\n\n```\n\nWhen specifying configuration in a .env file, include it in docker-compose so the configuration is passed into the container at runtime.\n\ndocker-compose.yml\n\n```yaml\nversion: "3.7"\n\nservices:\n  collector:\n    image: driftwood/fracx\n    env_file: .env\n```\n\n<br/>\n<br/>\n##### Using docker-compose.yml\n\nExample configuration directly in the docker-compose.yml file:\n\n```yaml\nservices:\n  collector:\n    image: driftwood/fracx\n    environment:\n      FRACX_FTP_USERNAME: YOUR_FRACX_FTP_UERNAME\n      FRACX_FTP_PASSWORD: YOUR_FRACX_FTP_PASSWORD\n      DATABASE_USERNAME: YOUR_DATABASE_USERNAME\n      DATABASE_PASSWORD: YOUR_DATABASE_PASSWORD\n      DATABASE_HOST: YOUR_DATABASE_HOST\n      DATABASE_NAME: YOUR_DATABASE_NAME\n      DATABASE_SCHEMA: fracx\n      FRAC_SCHEDULE_TABLE_NAME: frac_schedules\n```\n',
    'author': 'Brock Friedrich',
    'author_email': 'brocklfriedrich@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/la-mar/permian-frac-exchange',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import re


class StringProcessor:
    re_ws = re.compile(r"\s+")
    re_dup_ws = re.compile(r"\s\s+")
    re_non_alphanum = re.compile(r"(?ui)\W")
    re_non_num = re.compile(r"[^\d+]")
    replacement = "_"
    tolower = True

    def __init__(
        self, replacement: str = "_", tolower: bool = True, toupper: bool = False
    ):

        self.replacement = replacement
        self.tolower = tolower
        self.toupper = toupper

    def alphanum_only(self, s: str) -> str:
        """ Replace all non-alphanumeric characters with whitespace. Whitespace is forced as a replacement so separators can optionally be inserted later.
        """

        return self.re_non_alphanum.sub(" ", s)

    def numeric_only(self, s: str) -> str:
        """ Replace all non-numeric characters with whitespace. Whitespace is forced as a replacement so separators can optionally be inserted later.
        """

        return self.re_non_num.sub(" ", s)

    def dedupe_whitespace(self, s: str) -> str:
        """ remove duplicated whitespaces
        """
        return self.re_dup_ws.sub(" ", s)

    def remove_whitespace(self, s: str) -> str:
        """ remove all whitespaces
        """
        return self.re_ws.sub("", s)

    def fill_whitespace(self, s: str, replacement: str = None) -> str:
        replacement = replacement if replacement is not None else self.replacement
        return s.replace(" ", replacement)

    def normalize(self, s: str, int_compatable: bool = False) -> str:
        """Normalizes the given string. Operations performed on the string are:
                - remove all non-alphanumeric characters
                - squish sequential whitespace to a single space
                - convert to all lower case
                - strip leading and trailing whitespace

        Arguments:
            s {str} -- a string to process

        Keyword Arguments:
            int_compatable {bool} -- return an int parsable string. The string is NOT converted to an integer type.

        Returns:
            str -- the normalized string
        """

        if s is not None:
            s = str(s)
            s = self.alphanum_only(s)
            s = self.dedupe_whitespace(s)
            s = str.strip(s)

            if int_compatable:
                s = self.remove_whitespace(s)
                # dont actually convert to int to avoid the potential failure point during ingestion
            else:
                s = self.fill_whitespace(s)

                if self.toupper:
                    s = s.upper()
                elif self.tolower:
                    s = s.lower()

        return s



# flake8: noqa
from collector.downloader import Ftp, Sftp
from collector.endpoint import Endpoint
from collector.collector import Collector, FracScheduleCollector
from collector.transformer import Transformer
from collector.downloader import Ftp
from collector.filehandler import BytesFileHandler



# pylint: disable=unused-argument


import warnings

from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import exc as sa_exc

import loggers
from config import APP_SETTINGS, get_active_config


loggers.config()

conf = get_active_config()

# instantiate the extensions
db = SQLAlchemy()


# warnings.simplefilter("ignore", category=sa_exc.SAWarning)


def create_app(script_info=None):
    app = Flask(__name__)
    app.config.from_object(APP_SETTINGS)

    # set up extensions
    db.init_app(app)
    # db.reflect(app=app)

    # shell context for flask cli
    @app.shell_context_processor
    def ctx():
        return {"app": app, "db": db}

    return app



# pylint: disable=protected-access

""" Logger definitions """

import numbers
from typing import Any, Mapping, Union
import logging
import logging.config
import os
import sys
from logging import LogRecord

import json_log_formatter
import logutils.colorize


from config import get_active_config
from util.jsontools import ObjectEncoder


conf = get_active_config()


LOG_LEVELS = dict(logging._nameToLevel)
LOG_LEVELS.update(logging._levelToName)  # type: ignore
LOG_LEVELS.update({str(k): v for k, v in logging._levelToName.items()})  # type: ignore
LOG_LEVELS.setdefault("FATAL", logging.FATAL)
LOG_LEVELS.setdefault(logging.FATAL, "FATAL")  # type: ignore

logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("boto3").setLevel(logging.WARNING)
logging.getLogger("botocore").setLevel(logging.CRITICAL)
logging.getLogger("zeep").setLevel(logging.CRITICAL)
logging.getLogger("requests_oauthlib").setLevel(logging.CRITICAL)


def mlevel(level):
    """Convert level name/int to log level. Borrowed from celery, with <3"""
    if level and not isinstance(level, numbers.Integral):
        return LOG_LEVELS[level.upper()]
    return level


class ColorizingStreamHandler(logutils.colorize.ColorizingStreamHandler):
    """
    A stream handler which supports colorizing of console streams
    under Windows, Linux and Mac OS X.

    :param strm: The stream to colorize - typically ``sys.stdout``
                 or ``sys.stderr``.
    """

    # color names to indices
    color_map = {
        "black": 0,
        "red": 1,
        "green": 2,
        "yellow": 3,
        "blue": 4,
        "magenta": 5,
        "cyan": 6,
        "white": 7,
    }

    # levels to (background, foreground, bold/intense)
    if os.name == "nt":
        level_map = {
            logging.DEBUG: (None, "blue", True),
            logging.INFO: (None, "white", False),
            logging.WARNING: (None, "yellow", True),
            logging.ERROR: (None, "red", True),
            logging.CRITICAL: ("red", "white", True),
        }
    else:
        "Maps levels to colour/intensity settings."
        level_map = {
            logging.DEBUG: (None, "blue", False),
            logging.INFO: (None, "white", False),
            logging.WARNING: (None, "yellow", False),
            logging.ERROR: (None, "red", False),
            logging.CRITICAL: ("red", "white", True),
        }


class DatadogJSONFormatter(json_log_formatter.JSONFormatter):
    """JSON log formatter that includes Datadog standard attributes."""

    trace_enabled = False

    def format(self, record: LogRecord):
        """Return the record in the format usable by Datadog."""
        json_record = self.json_record(record.getMessage(), record)
        mutated_record = self.mutate_json_record(json_record)
        # Backwards compatibility: Functions that overwrite this but don't
        # return a new value will return None because they modified the
        # argument passed in.
        if mutated_record is None:
            mutated_record = json_record
        return self.to_json(mutated_record)

    def to_json(self, record: Mapping[str, Any]):
        """Convert record dict to a JSON string.
        Override this method to change the way dict is converted to JSON.
        """
        return self.json_lib.dumps(record, cls=ObjectEncoder)

    def json_record(self, message: str, record: LogRecord):
        """Convert the record to JSON and inject Datadog attributes."""
        record_dict = dict(record.__dict__)

        record_dict["message"] = message

        if "timestamp" not in record_dict:
            # UNIX time in milliseconds
            record_dict["timestamp"] = int(record.created * 1000)

        if "severity" not in record_dict:
            record_dict["severity"] = record.levelname

        # Source Code
        if "logger.name" not in record_dict:
            record_dict["logger.name"] = record.name
        if "logger.method_name" not in record_dict:
            record_dict["logger.method_name"] = record.funcName
        if "logger.thread_name" not in record_dict:
            record_dict["logger.thread_name"] = record.threadName

        # NOTE: We do not inject 'host', 'source', or 'service', as we want
        # Datadog agent and docker labels to handle that for the time being.
        # This may change.

        exc_info = record.exc_info
        try:
            if self.trace_enabled:
                # get correlation ids from current tracer context
                trace_id, span_id = [1, 2]  # helpers.get_correlation_ids()
                record_dict["dd.trace_id"] = trace_id or 0
                record_dict["dd.span_id"] = span_id or 0

            if "context" in record_dict:
                context_obj = dict()
                context_value = record_dict.get("context")
                if context_value:
                    array = context_value.replace(" ", "").split(",")
                else:
                    array = []
                for item in array:
                    key, val = item.split("=")

                    # del key from record before replacing with modified version
                    # NOTE: This is hacky. Need to provide a general purpose
                    # context-aware logger.
                    if key in record_dict:
                        del record_dict[key]

                    key = f"ctx.{key}"
                    context_obj[key] = int(val) if val.isdigit() else val
                    record_dict.update(context_obj)

                del record_dict["context"]
        except Exception:
            exc_info = sys.exc_info()

        # Handle exceptions, including those in our formatter
        if exc_info:
            # QUESTION: If exc_info was set by us, do we alter the log level?
            # Probably not, as a formatter should never be altering the record
            # directly.
            # I think that instead we should avoid code that can conveivably
            # raise exceptions in our formatter. That is not possible until we update
            # the context handling code and we can ensure helpers.get_correlation_ids()
            # will not raise any exceptions.
            if "error.kind" not in record_dict:
                record_dict["error.kind"] = exc_info[0].__name__  # type: ignore
            if "error.message" not in record_dict:
                record_dict["error.message"] = str(exc_info[1])
            if "error.stack" not in record_dict:
                record_dict["error.stack"] = self.formatException(exc_info)

        return record_dict


def get_formatter(name: Union[str, None]) -> logging.Formatter:
    formatters = {
        "verbose": logging.Formatter(
            fmt="[%(asctime)s - %(filename)s:%(lineno)s - %(funcName)s()] %(levelname)s - %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        ),
        "funcname": logging.Formatter(
            fmt="[%(name)s: %(lineno)s - %(funcName)s()] %(levelname)s - %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        ),
        "simple": logging.Formatter(
            fmt="%(name)s - %(levelname)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        ),
        "layman": logging.Formatter(fmt="%(message)s", datefmt="%Y-%m-%d %H:%M:%S"),
        "json": DatadogJSONFormatter(),
    }
    return formatters[name or "funcname"]  # type: ignore


def config(level: int = None, formatter: str = None, logger: logging.Logger = None):

    root_logger = logger or logging.getLogger()
    root_logger.setLevel(mlevel(conf.LOG_LEVEL) or level or 20)
    console_handler = ColorizingStreamHandler()
    console_handler.setFormatter(get_formatter(formatter or conf.LOG_FORMAT))
    if root_logger.handlers:
        root_logger.removeHandler(root_logger.handlers[0])
    root_logger.addHandler(console_handler)

    # file_handler = logging.FileHandler("log/log.log")
    # root_logger.addHandler(file_handler)
    root_logger.addHandler(console_handler)


if __name__ == "__main__":

    config()
    logger = logging.getLogger()
    logger.debug("test-debug")
    logger.info("test-info")
    logger.warning("test-warning")
    logger.error("test-error")
